"""GPU Implementation of upfirdn.

A separate implementation "convolve1d" is also provided as a somewhat faster
alternative in the case up=down=1 (i.e. standard convolution).
"""
from math import ceil

import numpy as np

import cupy
from cupy.util import memoize
from fast_upfirdn.cpu._upfirdn_apply import mode_enum as _get_mode_enum
from fast_upfirdn._util import profile


try:
    # Device Attributes require CuPy > 6.0.b3
    d = cupy.cuda.device.Device()
    cuda_MaxBlockDimX = d.attributes["MaxBlockDimX"]
    cuda_MaxGridDimX = d.attributes["MaxGridDimX"]
except (cupy.cuda.runtime.CUDARuntimeError, AttributeError):
    # guess
    cuda_MaxBlockDimX = 1024
    cuda_MaxGridDimX = 2147483647

__all__ = ["convolve1d", "upfirdn"]


def _output_len(len_h, in_len, up, down):
    """The output length that results from a given input.

    scipy.signal._upfirdn._output_len
    """
    in_len_copy = in_len + (len_h + (-len_h % up)) // up - 1
    nt = in_len_copy * up
    need = nt // down
    if nt % down > 0:
        need += 1
    return need


_include = r"""
#include <cupy/complex.cuh>

#if !defined(__CUDA_ARCH__) || __CUDA_ARCH__ >= 600
#else
    // atomicAdd for doubles didn't exist prior to compute capability 6.0
    __device__ double atomicAdd(double* address, double val)
    {
        unsigned long long int* address_as_ull =
                                 (unsigned long long int*)address;
        unsigned long long int old = *address_as_ull, assumed;
        do {
            assumed = old;
            old = atomicCAS(address_as_ull, assumed,
                            __double_as_longlong(val +
                                __longlong_as_double(assumed)));
        } while (assumed != old);
        return __longlong_as_double(old);
    }
#endif

// atomicAdd for complex floats via two real-valued atomicAdds

__device__ complex<float> atomicAdd(complex<float>* address,
                                    complex<float> val)
{
    float *p = reinterpret_cast<float *>(address);

    return complex<float>(atomicAdd(p, val.real()),
                          atomicAdd(p + 1, val.imag()));
}

__device__ complex<double> atomicAdd(complex<double>* address,
                                     complex<double> val)
{
    double *p = reinterpret_cast<double *>(address);

    return complex<double>(atomicAdd(p, val.real()),
                           atomicAdd(p + 1, val.imag()));
}

enum MODE {
    MODE_CONSTANT = 0,
    MODE_SYMMETRIC = 1,
    MODE_CONSTANT_EDGE = 2,
    MODE_SMOOTH = 3,
    MODE_PERIODIC = 4,
    MODE_REFLECT = 5,
    MODE_ANTISYMMETRIC = 6,
    MODE_ANTIREFLECT = 7,
    MODE_LINE = 8
};


__device__
DTYPE_DATA _extend_left(DTYPE_DATA *x, long long idx, long long len_x,
                        MODE mode, DTYPE_DATA cval)
{
    DTYPE_DATA le = 0.;
    DTYPE_DATA lin_slope = 0.;

    switch(mode)
    {
    // note: idx will be < 0
    case MODE_SYMMETRIC:
        if ((-idx) < len_x)
        {
            return x[-idx - 1];
        }
        else
        {
            // general case for multiple reflections:
            // the pattern repeats with periodicity 2*len_x;
            idx = (-idx - 1) % (2 * len_x);
            if (idx < len_x)
                return x[idx];
            else
                return x[len_x - 1 - (idx - len_x)];
        }
    case MODE_REFLECT:
        if ((-idx) < (len_x - 1))
        {
            return x[-idx];
        }
        else
        {
            // general case for multiple reflections:
            // the pattern repeats with periodicity 2*(len_x - 1);
            idx = (-idx - 1) % (2 * (len_x - 1));
            if (idx < (len_x - 1))
                return x[idx + 1];
            else
                return x[len_x - 2 - (idx - (len_x - 1))];
        }
    case MODE_PERIODIC:
        idx = (-idx - 1) % len_x;
        return x[len_x - idx - 1];
    case MODE_SMOOTH:
        return x[0] + (DTYPE_DATA)idx * (x[1] - x[0]);
    case MODE_LINE:
        lin_slope = (x[len_x - 1] - x[0]) / (DTYPE_DATA)(len_x - 1);
        return x[0] + (DTYPE_DATA)idx * lin_slope;
    case MODE_ANTISYMMETRIC:
        if ((-idx) < len_x)
        {
            return -x[-idx - 1];
        }
        else
        {
            idx = (-idx - 1) % (2 * len_x);
            if (idx < len_x)
            {
                return -x[idx];
            }
            else
            {
                return x[len_x - 1 - (idx - len_x)];
            }
        }
    case MODE_ANTIREFLECT:
        if ((-idx) < len_x)
        {
            return x[0] - (x[-idx] - x[0]);
        }
        else
        {
            le = x[0] + (x[0] - x[len_x - 1]) *
                 ((DTYPE_DATA)((-(idx) - 1) / (len_x - 1)));
            idx = (-idx - 1) % (2 * (len_x - 1));
            if (idx < (len_x - 1))
            {
                return le - (x[idx + 1] - x[0]);
            }
            else
            {
                return le - (
                    x[len_x - 1] - x[len_x - 2 - (idx - (len_x - 1))]);
            }
        }
    case MODE_CONSTANT_EDGE:
        return x[0];
    case MODE_CONSTANT:
        return cval;
    default:
        return -1.;
    }
}


__device__
DTYPE_DATA _extend_right(DTYPE_DATA *x, long long idx, long long len_x,
                           MODE mode, DTYPE_DATA cval)
{
    // note: idx will be >= len_x
    DTYPE_DATA re = 0.;
    DTYPE_DATA lin_slope = 0.;
    switch(mode)
    {

        case MODE_SYMMETRIC:
        {
            if (idx < (2 * len_x))
            {
                return x[len_x - 1 - (idx - len_x)];
            }
            else
            {
                idx = idx % (2 * len_x);
                if (idx < len_x)
                {
                    return x[idx];
                }
                else
                {
                    return x[len_x - 1 - (idx - len_x)];
                }
            }
        }
        case MODE_REFLECT:
        {
            if (idx < (2 * len_x - 1))
            {
                return x[len_x - 2 - (idx - len_x)];
            }
            else
            {
                idx = idx % (2 * (len_x - 1));
                if (idx < (len_x - 1))
                {
                    return x[idx];
                }
                else
                {
                    return x[len_x - 1 - (idx - (len_x - 1))];
                }
            }
        }
        case MODE_PERIODIC:
        {
            return x[idx % len_x];
        }
        case MODE_SMOOTH:
            return x[len_x - 1] +
                   (DTYPE_DATA)(idx - len_x + 1) *
                   (x[len_x - 1] - x[len_x - 2]);
        case MODE_LINE:
            lin_slope = (x[len_x - 1] - x[0]) / (DTYPE_DATA)(len_x - 1);
            return x[len_x - 1] + (DTYPE_DATA)(idx - len_x + 1) * lin_slope;
        case MODE_CONSTANT_EDGE:
            return x[len_x - 1];
        case MODE_ANTISYMMETRIC:
            if (idx < (2 * len_x))
            {
                return -x[len_x - 1 - (idx - len_x)];
            }
            else
            {
                idx = idx % (2 * len_x);
                if (idx < len_x)
                {
                    return x[idx];
                }
                else
                {
                    return -x[len_x - 1 - (idx - len_x)];
                }
            }
        case MODE_ANTIREFLECT:
            if (idx < (2 * len_x - 1))
            {
                return x[len_x - 1] - (
                    x[len_x - 2 - (idx - len_x)] - x[len_x - 1]);
            }
            else
            {
                re = x[len_x - 1] +
                     (x[len_x - 1] - x[0]) *
                     ((DTYPE_DATA)(idx / (len_x - 1) - 1));
                idx = idx % (2 * (len_x - 1));
                if (idx < (len_x - 1))
                {
                    return re + (x[idx] - x[0]);
                }
                else
                {
                    return re + (x[len_x - 1] -
                                 x[len_x - 1 - (idx - (len_x - 1))]);
                }
            }
        case MODE_CONSTANT:
            return cval;
        default:
            return -1.;
    }
}

"""


_convolved_batch_template = (
    _include
    + r"""

extern "C" {

__global__
void _apply_batch(DTYPE_DATA *x, long long len_x,
               DTYPE_FILTER *h_trans_flip,
               long long len_h,
               DTYPE_OUT *out,
               long long out_axis_size,
               long long nbatch,
               long long _mode,
               DTYPE_DATA cval,
               long long offset,
               long long crop)
{
    __shared__ DTYPE_FILTER h_trans_flip_s[128];
    long long x_conv_idx;
    long long i;
    // TODO: set initial values for these constants outside the loop
    long long unraveled_idx = blockDim.x * blockIdx.x + threadIdx.x;
    long long batch_idx = unraveled_idx / out_axis_size;
    MODE mode = (MODE)_mode;

    for (i=0; i<len_h; i++)
    {
        h_trans_flip_s[i] = h_trans_flip[i];
    }

    if (batch_idx < nbatch)
    {
        long long padded_len;
        long long offset_x = batch_idx * len_x;
        long long offset_out = batch_idx * out_axis_size;

        long long y_idx = unraveled_idx - offset_out;
        long long h_idx = 0;
        long long x_idx = y_idx + offset;

        DTYPE_OUT val = 0.0;
        DTYPE_OUT xval;

        bool zpad = (mode == MODE_CONSTANT) && (abs(cval) == 0.0);

        if (crop)
        {
            padded_len = len_x + offset;
        }
        else
        {
            padded_len = len_x + len_h - 1 + offset;
        }

        if (x_idx < len_x)
        {
            x_conv_idx = x_idx - len_h + 1;
            if (x_conv_idx < 0){
                if (zpad)
                {
                    h_idx -= x_conv_idx;
                }
                else
                {
                    for (; x_conv_idx < 0; x_conv_idx++){
                        xval = _extend_left(
                            &x[offset_x], x_conv_idx, len_x, mode, cval);
                        val += xval * h_trans_flip_s[h_idx];
                        h_idx++;
                    }
                }
                x_conv_idx = 0;
            }
            for (; x_conv_idx < x_idx + 1; x_conv_idx++){
                val += x[offset_x + x_conv_idx] * h_trans_flip_s[h_idx];
                h_idx++;
            }
            atomicAdd(&out[unraveled_idx], val);
        }

        // Use a second simplified loop to flush out the last bits
        else if (x_idx < padded_len)
        {
            x_conv_idx = x_idx - len_h + 1;
            for (; x_conv_idx < x_idx + 1; x_conv_idx++)
            {
                if (x_conv_idx >= len_x)
                {
                    xval = _extend_right(
                        &x[offset_x], x_conv_idx, len_x, mode, cval);

                }
                else if (x_conv_idx < 0)
                {
                    xval = _extend_left(
                        &x[offset_x], x_conv_idx, len_x, mode, cval);
                }
                else
                {
                    xval = x[offset_x + x_conv_idx];
                }
                val += xval * h_trans_flip_s[h_idx];
                h_idx++;
            }
            atomicAdd(&out[unraveled_idx], val);
        }
    }
}
}
"""
)


_upfirdn_part1_shared_h = r"""

extern "C" {

__global__
void _apply_batch(DTYPE_DATA *x, long long len_x,
                  DTYPE_FILTER *h_trans_flip,
                  long long len_h,
                  DTYPE_OUT *out,
                  long long up,
                  long long down,
                  long long out_axis_size,
                  long long nbatch,
                  long long _mode,
                  DTYPE_DATA cval,
                  long long offset,
                  long long crop)
{
    __shared__ DTYPE_FILTER h_trans_flip_s[128];
    long long x_conv_idx;
    long long i;
    // TODO: set initial values for these constants outside the loop
    long long unraveled_idx = blockDim.x * blockIdx.x + threadIdx.x;
    long long batch_idx = unraveled_idx / out_axis_size;
    MODE mode = (MODE)_mode;

    // copy filter to shared memory
    for (i=0; i<len_h; i++)
    {
       h_trans_flip_s[i] = h_trans_flip[i];
    }
"""

_upfirdn_part1_nonshared_h = r"""

extern "C" {

__global__
void _apply_batch(DTYPE_DATA *x, long long len_x,
                  DTYPE_FILTER *h_trans_flip_s,
                  long long len_h,
                  DTYPE_OUT *out,
                  long long up,
                  long long down,
                  long long out_axis_size,
                  long long nbatch,
                  long long _mode,
                  DTYPE_DATA cval,
                  long long offset,
                  long long crop)
{
    long long x_conv_idx;
    long long i;
    // TODO: set initial values for these constants outside the loop
    long long unraveled_idx = blockDim.x * blockIdx.x + threadIdx.x;
    long long batch_idx = unraveled_idx / out_axis_size;
    MODE mode = (MODE)_mode;

"""

_upfirdn_template_part2 = r"""

    if (batch_idx < nbatch)
    {
        long long h_per_phase = len_h / up;
        long long padded_len;
        long long offset_x = batch_idx * len_x;
        long long offset_out = batch_idx * out_axis_size;

        long long y_idx = unraveled_idx - offset_out;
        long long t = ((y_idx + offset)*down) % up;
        long long h_idx = t * h_per_phase;
        long long x_idx = ((y_idx + offset)*down) / up;

        DTYPE_OUT val = 0.0;
        DTYPE_OUT xval;
        DTYPE_OUT hval = 0.0;

        bool zpad = (mode == MODE_CONSTANT) && (abs(cval) == 0.0);
        if (crop)
            padded_len = len_x + offset * down / up;
        else
            padded_len = len_x + len_h - 1 + offset * down / up;

        if (x_idx < len_x)
        {
            x_conv_idx = x_idx - h_per_phase + 1;
            if (x_conv_idx < 0){
                if (zpad)
                {
                    h_idx -= x_conv_idx;
                }
                else
                {
                    for (; x_conv_idx < 0; x_conv_idx++){
                        hval = h_trans_flip_s[h_idx];
                        if (hval != DTYPE_FILTER(0))
                        {
                            xval = _extend_left(
                                &x[offset_x], x_conv_idx, len_x, mode, cval);
                            val += xval * hval;
                        }
                        h_idx++;
                    }
                }
                x_conv_idx = 0;
            }
            for (; x_conv_idx < x_idx + 1; x_conv_idx++){
                hval = h_trans_flip_s[h_idx];
                if (hval != DTYPE_FILTER(0))
                {
                    val += x[offset_x + x_conv_idx] * hval;
                }
                h_idx++;
            }
            atomicAdd(&out[unraveled_idx], val);
        }

        // Use a second simplified loop to flush out the last bits
        else if (x_idx < padded_len)
        {
            x_conv_idx = x_idx - h_per_phase + 1;
            for (; x_conv_idx < x_idx + 1; x_conv_idx++)
            {
                hval = h_trans_flip_s[h_idx];
                if (hval != DTYPE_FILTER(0))
                {
                    if (x_conv_idx >= len_x)
                    {
                        xval = _extend_right(
                            &x[offset_x], x_conv_idx, len_x, mode, cval);

                    }
                    else if (x_conv_idx < 0)
                    {
                        xval = _extend_left(
                            &x[offset_x], x_conv_idx, len_x, mode, cval);
                    }
                    else
                    {
                        xval = x[offset_x + x_conv_idx];
                    }
                    val += xval * hval;
                }
                h_idx++;
            }
            atomicAdd(&out[unraveled_idx], val);
        }
    }
}
}
"""

# version where the filter, h, is copied into local shared memory
_upfirdn_batch_template_shared_h = (
    _include + _upfirdn_part1_shared_h + _upfirdn_template_part2
)

# version where the filter, h, is not copied into local shared memory
_upfirdn_batch_template_nonshared_h = (
    _include + _upfirdn_part1_nonshared_h + _upfirdn_template_part2
)


# dictionary: CUDA C data types corresponding to numpy dtype.char values
c_dtypes = {
    "f": "float",
    "d": "double",
    "F": "complex<float>",
    "D": "complex<double>",
}


@memoize()
def _nearest_supported_float_dtype(dtype, dtype2=None):
    if dtype.char in ["f", "d", "F", "D"] and (
        dtype2 is None or dtype2 == dtype
    ):
        return dtype, c_dtypes.get(dtype.char)

    # determine nearest single or double precision floating point type
    dtype = np.promote_types(dtype, np.float32)
    if dtype2 is not None:
        dtype = np.promote_types(dtype, dtype2)
    if dtype.char == "g":
        dtype = np.dtype(np.float64)
    elif dtype.char == "G":
        dtype = np.dtype(np.complex128)
    return dtype, c_dtypes.get(dtype.char)


@memoize(for_each_device=True)
def _get_upfirdn_kernel_inner(
    up, down, c_dtype_data, c_dtype_filter, c_dtype_out, h_size
):
    from cupy.cuda import nvrtc

    func_name = "_apply_batch"

    # crude template-like functionality via string replacement
    if up == down == 1:
        code = _convolved_batch_template.replace("DTYPE_DATA", c_dtype_data)
    else:
        if h_size <= 128:
            code = _upfirdn_batch_template_shared_h.replace(
                "DTYPE_DATA", c_dtype_data
            )
        else:
            code = _upfirdn_batch_template_nonshared_h.replace(
                "DTYPE_DATA", c_dtype_data
            )

    code = code.replace("DTYPE_FILTER", c_dtype_filter)
    code = code.replace("DTYPE_OUT", c_dtype_out)

    if nvrtc.getVersion() < (9, 2):
        # __shared__ complex<T> doesn't work on older CUDA compilers

        """ From the CUDA 9.2 release notes:
        The CUDA compiler previously incorrectly determined that the
        constructor for a __shared__ multidimensional array variable was
        non-empty in some scenarios, and generated a spurious diagnostic.
        The bug has now been fixed.
        """
        code = code.replace("__shared__ complex", "complex")

    kern = cupy.RawKernel(code, func_name)
    return kern


@memoize()
def _determine_dtypes(data_dtype, data_real_dtype, h_dtype, h_real_dtype):
    # note need to include h_real_dtype here so data's dtype will be promoted
    # if h has a higher precision dtype.
    dtype_data, c_dtype_data = _nearest_supported_float_dtype(
        data_dtype, h_real_dtype
    )
    if data_dtype == h_dtype:
        dtype_out = dtype_filter = dtype_data
        c_dtype_out = c_dtype_filter = c_dtype_data
    else:
        # convert h to the same precision as data if there is a mismatch
        cplx_h = h_dtype.kind == "c"
        if data_real_dtype != h_real_dtype:
            if cplx_h:
                h_dtype = np.promote_types(dtype_data, np.complex64)
            else:
                h_dtype = np.promote_types(dtype_data, np.float32)

        dtype_filter, c_dtype_filter = _nearest_supported_float_dtype(h_dtype)
        if cplx_h:
            # output is complex if filter is complex
            c_dtype_out = c_dtype_filter
            dtype_out = dtype_filter
        else:
            # for real filter, output dtype matches the data dtype
            c_dtype_out = c_dtype_data
            dtype_out = dtype_data
    return (
        dtype_data,
        c_dtype_data,
        dtype_filter,
        c_dtype_filter,
        dtype_out,
        c_dtype_out,
    )


@profile
def get_upfirdn_kernel(h, data, up, down):
    """Compile an upfirdn kernel based on dtype.

    Also converts h, data to the nearest supported floating point type.
    """
    dt_data, c_dt_data, dt_h, c_dt_h, dt_out, c_dt_out = _determine_dtypes(
        data.dtype, data.real.dtype, h.dtype, h.real.dtype
    )

    if data.dtype != dt_data:
        data = data.astype(dt_data)

    if h.dtype != dt_h:
        h = h.astype(dt_h)

    # memoized GPU kernels
    kern = _get_upfirdn_kernel_inner(
        up, down, c_dt_data, c_dt_h, c_dt_out, h.size
    )

    return h, data, dt_out, kern


def _pad_h(h, up):
    """Store coefficients in a transposed, flipped arrangement.

    For example, suppose upRate is 3, and the
    input number of coefficients is 10, represented as h[0], ..., h[9].

    Then the internal buffer will look like this::

       h[9], h[6], h[3], h[0],   // flipped phase 0 coefs
       0,    h[7], h[4], h[1],   // flipped phase 1 coefs (zero-padded)
       0,    h[8], h[5], h[2],   // flipped phase 2 coefs (zero-padded)

    Notes
    -----
    This is a copy of the function from scipy.signal._upfirdn.py with cupy
    support.
    """
    if up == 1:
        return h[::-1].copy()  # copy to avoid negative strides
    else:
        lh = len(h)
        h_padlen = lh + (-lh % up)
        if h_padlen == 0:
            h_full = h
        else:
            h_full = cupy.zeros(h_padlen, h.dtype)
            h_full[:lh] = h
        h_full = h_full.reshape(-1, up).T[:, ::-1].ravel()
        # TODO: numpy doesn't need this ascontiguousarray here,
        #       but cupy does to avoid negative strides!
        h_full = cupy.ascontiguousarray(h_full)
    return h_full


def convolve1d(
    h,
    x,
    axis=-1,
    contiguous_output=False,
    block_size=32,
    out=None,
    mode="zero",
    cval=0,
    offset=0,
    crop=False,
):
    """

    out : use a preallocated output array.
        Unless filtering is being performed along the last axis, a copy may
        have to be made. For this reason, one should not rely on operation
        upon the ``out`` array to be in-place.
    """

    # compile or retrieve cached kernel for the given dtypes
    h, x, dtype_out, kern = get_upfirdn_kernel(h, x, up=1, down=1)

    ndim = x.ndim

    mode_enum = _get_mode_enum(mode)
    cval = dtype_out.type(cval)

    # flip the filter
    h_flip = h[::-1].copy()  # copy to avoid negative strides

    len_h = len(h_flip)
    if axis < -ndim or axis > ndim - 1:
        raise ValueError("axis out of range")
    axis = axis % x.ndim
    if axis != ndim - 1:
        x = x.swapaxes(axis, -1)
    out_shape = [s for s in x.shape]
    if not crop:
        out_len = x.shape[-1] + len_h - 1
    else:
        out_len = x.shape[-1]
    out_shape[-1] = out_len

    if out is not None and out is x:
        # avoid clobbering existing values in x
        x = x.copy()
    else:
        x = cupy.ascontiguousarray(x)
    x = x.reshape((-1, x.shape[-1]), order="C")
    nbatch = x.shape[0]

    inplace_output = False
    if out is None:
        y = cupy.zeros((nbatch, out_len), dtype=dtype_out)
        out = y
    else:
        _out_orig = out
        # output into preallocated array
        if axis != ndim - 1:
            out = out.swapaxes(-1, axis)
        if out.size != nbatch * out_len or out.shape[-1] != out_len:
            raise ValueError("out array has the wrong size")
        elif out.dtype != dtype_out:
            raise ValueError(
                "Expected an out array with dtype: {}".format(dtype_out)
            )
        if not out.flags.c_contiguous:
            out = cupy.ascontiguousarray(out)
        out[:] = 0.0
        y = out
        inplace_output = True

    grid_size_x = ceil(y.size / block_size)
    if grid_size_x > cuda_MaxGridDimX:
        raise ValueError(
            "Grid size > MaxGridDimX for the GPU. Try increasing block_size."
        )
    if block_size > cuda_MaxBlockDimX:
        raise ValueError("block_size exceeds MaxBlockDimX for the GPU")

    kern(
        (grid_size_x,),
        (block_size,),
        (
            x,
            x.shape[-1],
            h_flip,
            len_h,
            y,
            out_len,
            nbatch,
            mode_enum,
            cval,
            offset,
            crop,
        ),
    )
    y = y.reshape(out_shape, order="C")

    if axis != ndim - 1:
        y = y.swapaxes(axis, -1)
    if contiguous_output:
        y = cupy.ascontiguousarray(y)
    if inplace_output:
        if _out_orig is not y:
            if _out_orig.shape != y.shape:
                raise ValueError(
                    "output array does not have the expected shape"
                )
            _out_orig[...] = y[...]
        return _out_orig
    return y


@profile
def upfirdn(
    h,
    x,
    up=1,
    down=1,
    axis=-1,
    mode="zero",
    cval=0,
    *,
    contiguous_output=False,
    block_size=32,
    prepadded=False,
    out=None,
    crop=False,
    take=None,
    offset=0,
):
    """

    out : use a preallocated output array.
        Unless filtering is being performed along the last axis, a copy may
        have to be made. For this reason, one should not rely on operation
        upon the ``out`` array to be in-place.
    """
    if not isinstance(h, cupy.ndarray):
        h = cupy.asarray(h)
    if not isinstance(x, cupy.ndarray):
        x = cupy.asarray(x)

    # dev = cupy.cuda.Device()

    if down < 1 or up < 1:
        raise ValueError("Both up and down must be >= 1")
    if h.ndim != 1 or h.size == 0:
        raise ValueError("h must be 1-D with non-zero length")

    # compile or retrieve cached kernel for the given dtypes
    h, x, dtype_out, kern = get_upfirdn_kernel(h, x, up=up, down=down)

    ndim = x.ndim

    mode_enum = _get_mode_enum(mode)
    cval = dtype_out.type(cval)

    # flip the filter
    if prepadded:
        h_flip = h
    else:
        h_flip = _pad_h(h, up=up)

    len_h = len(h_flip)

    if axis < -ndim or axis > ndim - 1:
        raise ValueError("axis out of range")
    axis = axis % x.ndim
    if axis != ndim - 1:
        x = x.swapaxes(axis, -1)
    out_shape = [s for s in x.shape]
    if crop:
        out_len = int(ceil(x.shape[-1] * up / down))
    else:
        out_len = _output_len(h_flip.size, x.shape[-1], up, down)
    out_shape[-1] = out_len

    if out is not None and out is x:
        # avoid clobbering existing values in x
        x = x.copy()
    else:
        x = cupy.ascontiguousarray(x)
    x = x.reshape((-1, x.shape[-1]), order="C")
    nbatch = x.shape[0]

    inplace_output = False
    if out is None:
        y = cupy.zeros((nbatch, out_len), dtype=dtype_out)
        out = y
    else:
        _out_orig = out
        # output into preallocated array
        if axis != ndim - 1:
            out = out.swapaxes(-1, axis)
        if out.size != nbatch * out_len or out.shape[-1] != out_len:
            raise ValueError("out array has the wrong size")
        elif out.dtype != dtype_out:
            raise ValueError(
                "Expected an out array with dtype: {}".format(dtype_out)
            )
        if not out.flags.c_contiguous:
            out = cupy.ascontiguousarray(out)
        out[:] = 0.0
        y = out
        inplace_output = True

    grid_size_x = ceil(y.size / block_size)
    if grid_size_x > cuda_MaxGridDimX:
        raise ValueError(
            "Grid size > MaxGridDimX for the GPU. Try increasing block_size."
        )
    if block_size > cuda_MaxBlockDimX:
        raise ValueError("block_size exceeds MaxBlockDimX for the GPU")
    if up == down == 1:
        kern(
            (grid_size_x,),
            (block_size,),
            (
                x,
                x.shape[-1],
                h_flip,
                len_h,
                y,
                out_len,
                nbatch,
                mode_enum,
                cval,
                offset,
                int(crop),
            ),
        )
    else:
        kern(
            (grid_size_x,),
            (block_size,),
            (
                x,
                x.shape[-1],
                h_flip,
                len_h,
                y,
                up,
                down,
                out_len,
                nbatch,
                mode_enum,
                cval,
                offset,
                int(crop),
            ),
        )
    y = y.reshape(out_shape, order="C")

    if take is not None:
        # TODO: move into the kernel
        y_sl = [slice(None)] * y.ndim
        if take is None:
            y_sl[-1] = slice(None)
        else:
            if isinstance(take, slice):
                y_sl[-1] = take
            else:
                y_sl[-1] = slice(take)
        y = y[tuple(y_sl)]

    if axis != ndim - 1:
        y = y.swapaxes(axis, -1)
    if contiguous_output:
        y = cupy.ascontiguousarray(y)
    if inplace_output:
        if _out_orig is not y:
            if _out_orig.shape != y.shape:
                raise ValueError(
                    "output array does not have the expected shape"
                )
            _out_orig[...] = y[...]
        return _out_orig
    return y



"""Implementation of functions from the scipy.signal API.

Currently this is just ``upfirdn`` and ``resample_poly``.

Also defines ``upfirdn_out_len`` which is not part of the public scipy API.

"""
from functools import partial

from fast_upfirdn.cpu import upfirdn as upfirdn_cpu
from fast_upfirdn.cpu._upfirdn import _upfirdn_modes as upfirdn_modes
from fast_upfirdn.cpu._upfirdn_apply import _output_len as upfirdn_out_len
from fast_upfirdn._util import (
    get_array_module,
    have_cupy,
    check_device,
    profile,
)

if have_cupy:
    from fast_upfirdn.cupy import upfirdn as upfirdn_cupy


__all__ = ["upfirdn", "upfirdn_out_len", "upfirdn_modes"]


@profile
def upfirdn(
    h,
    x,
    up=1,
    down=1,
    axis=-1,
    mode="zero",
    cval=0,
    *,
    xp=None,
    prepadded=False,
    out=None,
    offset=0,
    crop=False,
    take=None,
):
    """

    See ``scipy.signal.upfirdn``. This version supports some additional
    keyword arguments.

    Additional Parameters
    ---------------------
    xp : module or None
        The array module (``cupy`` or ``numpy``). If not provided, it is
        inferred from the type of ``x``.
    prepadded : bool, optional
        If this is True, it is assumed that the internal computation
        ``h = _pad_h(h, up=up)`` has already been performed on ``h``.
    out : ndarray
        TODO
    mode : str, optional
        The signal extension mode used at the boundaries.
    cval : float, optional
        The constant value used when ``mode == "constant"``.
    offset : int, optional
        TODO
    crop : bool, optional
        TODO
    take : int or None, optional
        TODO
    """

    # If xp was unspecified, determine it from x
    xp, on_gpu = get_array_module(x, xp)

    # make sure both arrays are on the CPU when xp==numpy or GPU when xp==cupy
    x, h = map(partial(check_device, xp=xp), [x, h])

    upfirdn_kwargs = dict(
        up=up,
        down=down,
        axis=axis,
        mode=mode,
        cval=cval,
        offset=offset,
        crop=int(crop),
        take=take,
    )

    if on_gpu:
        y = upfirdn_cupy(h, x, prepadded=prepadded, out=out, **upfirdn_kwargs)
    else:
        if prepadded:
            raise ValueError("prepadded not supported on the CPU")
        if out is not None:
            raise ValueError("preallocated out array not supported on the CPU")
        y = upfirdn_cpu(h, x, **upfirdn_kwargs)
    return y



from timeit import timeit

setup = """
import cupy
import numpy as np
from scipy.signal import upfirdn as upfirdn_scipy
from fast_upfirdn import upfirdn
from scipy import ndimage as ndi

d = cupy.cuda.device.Device()
order = 'C'
shape = (192, 192, 192)
down = 2
up = 1
dtype_data = dtype_filter = np.float32
size = int(np.prod(shape))
x_cpu = np.arange(size, dtype=dtype_data).reshape(shape, order=order)
h_cpu = np.arange(7, dtype=dtype_filter)
x = cupy.asarray(x_cpu, order=order)
h = cupy.asarray(h_cpu)
"""

# warm start to avoid overhead from imports and initial kernel compilation
timeit(
    "upfirdn_scipy(h_cpu, x_cpu, up=up, down=down, axis=-1)",
    setup=setup,
    number=1,
)
timeit(
    "upfirdn_scipy(h_cpu, x_cpu, up=up, down=down, axis=0)",
    setup=setup,
    number=1,
)
timeit("upfirdn(h, x, up=up, down=down, axis=0)", setup=setup, number=1)
timeit("upfirdn(h, x, up=up, down=down, axis=-1)", setup=setup, number=1)

nreps = 100

t_cpu_cont = (
    timeit(
        "upfirdn_scipy(h_cpu, x_cpu, up=up, down=down, axis=-1)",
        setup=setup,
        number=nreps,
    )
    / nreps
)
print("Duration (CPU, contiguous axis) = {} ms".format(1000 * t_cpu_cont))

t_cpu_noncont = (
    timeit(
        "upfirdn_scipy(h_cpu, x_cpu, up=up, down=down, axis=0)",
        setup=setup,
        number=nreps,
    )
    / nreps
)
print(
    "Duration (CPU, non-contiguous axis) = {} ms".format(1000 * t_cpu_noncont)
)

t_gpu_cont = (
    timeit(
        "upfirdn(h, x, up=up, down=down, axis=-1); d.synchronize()",
        setup=setup,
        number=nreps,
    )
    / nreps
)
print("Duration (GPU, contiguous axis) = {} ms".format(1000 * t_gpu_cont))

t_gpu_noncont = (
    timeit(
        "upfirdn(h, x, up=up, down=down, axis=0); d.synchronize()",
        setup=setup,
        number=nreps,
    )
    / nreps
)
print(
    "Duration (GPU, non-contiguous axis) = {} ms".format(1000 * t_gpu_noncont)
)



from os.path import join as pjoin

# Format expected by setup.py and doc/source/conf.py: string of form "X.Y.Z"
_version_major = 0
_version_minor = 2
_version_micro = ""  # use "" for first of series, number for 1 and above
_version_extra = ""  # use "dev0" for developemnt, "" for full release

# Construct full version string from these.
_ver = [_version_major, _version_minor]
if _version_micro:
    _ver.append(_version_micro)
if _version_extra:
    _ver.append(_version_extra)

__version__ = ".".join(map(str, _ver))

CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: BSD License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Topic :: Scientific/Engineering",
]

# Description should be a one-liner:
description = "fast_upfirdn: CPU & GPU implementations of scipy.signal.upfirdn"
# Long description will go up on the pypi page
long_description = """

Fast Upfirdn
============
The core low-level function implemented here is an equivalent of
``scipy.signal.upfirdn`` but with support for both CPU (via [NumPy]) and GPU
(via [CuPy]).

The version of ``upfirdn`` here supports several signal extension modes. These
have been contributed upstream to SciPy and are available there for SciPy 1.4+.

.. _README: https://github.com/mritools/fast_upfirdn/blob/master/README.md

License
=======
``fast_upfirdn`` is licensed under the terms of the BSD 3-clause license. See
the file "LICENSE.txt" for information on the history of this software, terms &
conditions for usage, and a DISCLAIMER OF ALL WARRANTIES.

All trademarks referenced herein are property of their respective holders.

Copyright (c) 2019-2020,
Gregory R. Lee, Cincinnati Children's Hospital Medical Center.
"""

NAME = "fast_upfirdn"
MAINTAINER = "Gregory R. Lee"
MAINTAINER_EMAIL = "grlee77@gmail.com"
DESCRIPTION = description
LONG_DESCRIPTION = long_description
URL = "http://github.com/mritools/fast_upfirdn"
DOWNLOAD_URL = ""
LICENSE = "BSD"
AUTHOR = "Gregory R. Lee"
AUTHOR_EMAIL = "grlee77@gmail.com"
PLATFORMS = "OS Independent"
MAJOR = _version_major
MINOR = _version_minor
MICRO = _version_micro
VERSION = __version__
PACKAGE_DATA = {"fast_upfirdn": [pjoin("data", "*"), pjoin("tests", "*")]}
REQUIRES = ["numpy", "cython"]
PYTHON_REQUIRES = ">= 3.6"



from itertools import product

import numpy as np
import pytest

from scipy.signal import upfirdn as upfirdn_scipy
from fast_upfirdn import upfirdn
from fast_upfirdn.cpu._upfirdn_apply import _pad_test
from fast_upfirdn.cpu._upfirdn import _upfirdn_modes


@pytest.mark.parametrize(
    "dtype_data, dtype_filter",
    product(
        [
            np.float32,
            np.float64,
            np.complex64,
            np.complex128,
            np.float16,
            np.uint8,
            np.int16,
            np.int32,
            int,
        ],
        [np.float32, np.float64, np.complex64, np.complex128],
    ),
)
def test_dtype_combos_cpu(dtype_data, dtype_filter):
    shape = (64, 64)
    size = int(np.prod(shape))
    x = np.arange(size, dtype=dtype_data).reshape(shape)
    h = np.arange(5, dtype=dtype_filter)

    # up=1 kernel case
    np.testing.assert_allclose(
        upfirdn_scipy(h, x, up=1, down=2), upfirdn(h, x, up=1, down=2)
    )

    # down=1 kernel case
    np.testing.assert_allclose(
        upfirdn_scipy(h, x, up=2, down=1), upfirdn(h, x, up=2, down=1)
    )


@pytest.mark.parametrize(
    "nh, nx", product([2, 3, 4, 5, 6, 7, 8], [16, 17, 18, 19, 20])
)
def test_input_and_filter_sizes_cpu(nh, nx):
    dtype_data = dtype_filter = np.float32
    x = np.arange(nx, dtype=dtype_data)
    h = np.arange(1, nh + 1, dtype=dtype_filter)

    # up=1 kernel case
    np.testing.assert_allclose(
        upfirdn_scipy(h, x, up=1, down=2), upfirdn(h, x, up=1, down=2)
    )

    # down=1 kernel case
    np.testing.assert_allclose(
        upfirdn_scipy(h, x, up=2, down=1), upfirdn(h, x, up=2, down=1)
    )


@pytest.mark.parametrize(
    "shape, axis, order",
    product(
        [(16, 8), (24, 16, 8), (8, 9, 10, 11)],
        [0, 1, 2, 3, -1],  # 0, 1, -1],
        ["C", "F"],
    ),
)
def test_axis_and_order_cpu(shape, axis, order):
    dtype_data = dtype_filter = np.float32
    size = int(np.prod(shape))
    x = np.arange(size, dtype=dtype_data).reshape(shape, order=order)
    h = np.arange(3, dtype=dtype_filter)
    ndim = len(shape)
    if axis >= -ndim and axis < ndim:
        # up=1 case
        np.testing.assert_allclose(
            upfirdn_scipy(h, x, up=1, down=2, axis=axis),
            upfirdn(h, x, up=1, down=2, axis=axis),
        )

        # down=1 case
        np.testing.assert_allclose(
            upfirdn_scipy(h, x, up=2, down=1, axis=axis),
            upfirdn(h, x, up=2, down=1, axis=axis),
        )
    else:
        with pytest.raises(ValueError):
            upfirdn(h, x, up=2, down=1, axis=axis)


@pytest.mark.parametrize(
    "up, down, nx, nh",
    product([1, 2, 3, 4, 5], [1, 2, 3, 4, 5], [16, 17], [3, 4]),
)
def test_general_up_and_down_cpu(up, down, nx, nh):
    dtype_data = dtype_filter = np.float32
    x = np.arange(nx, dtype=dtype_data)
    h = np.arange(1, nh + 1, dtype=dtype_filter)

    np.testing.assert_allclose(
        upfirdn_scipy(h, x, up=up, down=down), upfirdn(h, x, up=up, down=down)
    )


@pytest.mark.parametrize("mode", _upfirdn_modes)
def test_extension_modes(mode):
    """Test vs. manually computed results for modes not in numpy's pad."""
    x = np.array([1, 2, 3, 1], dtype=float)
    npre, npost = 6, 6
    y = _pad_test(x, npre=npre, npost=npost, mode=mode)
    if mode == "antisymmetric":
        y_expected = np.asarray(
            [3, 1, -1, -3, -2, -1, 1, 2, 3, 1, -1, -3, -2, -1, 1, 2]
        )
    elif mode == "antireflect":
        y_expected = np.asarray(
            [1, 2, 3, 1, -1, 0, 1, 2, 3, 1, -1, 0, 1, 2, 3, 1]
        )
    elif mode == "smooth":
        y_expected = np.asarray(
            [-5, -4, -3, -2, -1, 0, 1, 2, 3, 1, -1, -3, -5, -7, -9, -11]
        )
    elif mode == "line":
        lin_slope = (x[-1] - x[0]) / (len(x) - 1)
        left = x[0] + np.arange(-npre, 0, 1) * lin_slope
        right = x[-1] + np.arange(1, npost + 1) * lin_slope
        y_expected = np.concatenate((left, x, right))
    else:
        y_expected = np.pad(x, (npre, npost), mode=mode)
    np.testing.assert_allclose(y, y_expected)


@pytest.mark.parametrize("mode", _upfirdn_modes)
def test_extension_modes_via_convolve(mode):
    """Test vs. manually computed results for modes not in numpy's pad."""
    x = np.array([1, 2, 3, 1], dtype=float)
    npre, npost = 6, 6
    # use impulse response filter to probe values extending past the original
    # array boundaries
    h = np.zeros((npre + 1 + npost,), dtype=float)
    h[npre] = 1

    if mode == "constant":
        cval = 5.0
        y = upfirdn(h, x, up=1, down=1, mode=mode, cval=cval)
    else:
        y = upfirdn(h, x, up=1, down=1, mode=mode)

    if mode == "antisymmetric":
        y_expected = np.asarray(
            [3, 1, -1, -3, -2, -1, 1, 2, 3, 1, -1, -3, -2, -1, 1, 2]
        )
    elif mode == "antireflect":
        y_expected = np.asarray(
            [1, 2, 3, 1, -1, 0, 1, 2, 3, 1, -1, 0, 1, 2, 3, 1]
        )
    elif mode == "smooth":
        y_expected = np.asarray(
            [-5, -4, -3, -2, -1, 0, 1, 2, 3, 1, -1, -3, -5, -7, -9, -11]
        )
    elif mode == "line":
        lin_slope = (x[-1] - x[0]) / (len(x) - 1)
        left = x[0] + np.arange(-npre, 0, 1) * lin_slope
        right = x[-1] + np.arange(1, npost + 1) * lin_slope
        y_expected = np.concatenate((left, x, right))
    elif mode == "constant":
        y_expected = np.pad(x, (npre, npost), mode=mode, constant_values=cval)
    else:
        y_expected = np.pad(x, (npre, npost), mode=mode)
    y_expected = y_expected.astype(y.dtype)
    np.testing.assert_allclose(y, y_expected)



"""Miscellaneous utility functions."""

import numpy as np

try:
    import cupy

    have_cupy = True
except ImportError:
    have_cupy = False

allow_device_transfers = False

"""
@profile decorator that does nothing when line_profiler is not active

see:
http://stackoverflow.com/questions/18229628/python-profiling-using-line-profiler-clever-way-to-remove-profile-statements
"""
try:
    import builtins

    profile = builtins.__dict__["profile"]
except (AttributeError, KeyError):

    def profile(func):
        """No line profiler, provide a pass-through version."""
        return func


__all__ = [
    "allow_device_transfers",
    "have_cupy",
    "get_array_module",
    "array_on_device",
    "check_device",
]


def get_array_module(arr, xp=None):
    """ Check if the array is a cupy GPU array and return the array module.

    Parameters
    ----------
    arr : numpy.ndarray or cupy.core.core.ndarray
        The array to check.

    Returns
    -------
    array_module : python module
        This will be cupy when on_gpu is True and numpy otherwise.
    on_gpu : bool
        Boolean indicating whether the array is on the GPU.
    """
    if xp is None:
        if isinstance(arr, np.ndarray) or not have_cupy:
            return np, False
        else:
            xp = cupy.get_array_module(arr)
            return xp, (xp != np)
    else:
        return xp, (xp != np)


def check_device(arr, xp, autoconvert=allow_device_transfers):
    if autoconvert:
        return array_on_device(arr, xp)
    elif xp == np:
        if isinstance(arr, np.ndarray):
            return arr
        elif hasattr(arr, "__array_interface__") and not (
            have_cupy and hasattr(arr, "__cuda_array_interface__")
        ):
            return np.asarray(arr)
        else:
            raise ValueError(
                "Expected a numpy.ndarray, got {}".format(type(arr))
            )
    elif have_cupy and xp == cupy:
        if isinstance(arr, cupy.ndarray):
            return arr
        elif hasattr(arr, "__cuda_array_interface__"):
            return cupy.asarray(arr)
        else:
            raise ValueError(
                "Expected a cupy.ndarray, got {}".format(type(arr))
            )
    else:
        raise ValueError("unsupported module: {}".format(xp))


def array_on_device(arr, xp):
    """Transfer arr to device corresponding to xp.

    Paramters
    ---------
    arr : np.ndarray or cupy.ndarray
        The array.
    xp : {numpy, cupy}
        The desired ndarray module.

    Returns
    -------
    A cupy.ndarray if ``xp == cupy``, otherwise a numpy.ndarray.
    """
    if xp == np:
        if have_cupy and hasattr(arr, "__cuda_array_interface__"):
            # copy back from GPU
            return arr.get()
    return xp.asarray(arr)



#!/usr/bin/env python
import os
import sys
from setuptools import setup, find_packages
from distutils.extension import Extension
from Cython.Distutils import build_ext
import numpy

PACKAGES = find_packages()

# Get version and release info, which is all stored in fast_upfirdn/version.py
ver_file = os.path.join("fast_upfirdn", "version.py")
with open(ver_file) as f:
    exec(f.read())
# Give setuptools a hint to complain if it's too old a version
# 24.2.0 added the python_requires option
# Should match pyproject.toml
SETUP_REQUIRES = ["setuptools >= 24.2.0"]
# This enables setuptools to install wheel on-the-fly
SETUP_REQUIRES += ["wheel"] if "bdist_wheel" in sys.argv else []

try:
    from Cython.Build import cythonize
except ImportError:
    USE_CYTHON = False
else:
    USE_CYTHON = True

# Obtain the numpy include directory.  This logic works across numpy versions.
try:
    numpy_include = numpy.get_include()
except AttributeError:
    numpy_include = numpy.get_numpy_include()

# cython check
try:
    import cython

    # check that cython version is > 0.23
    cython_version = cython.__version__
    if float(cython_version.partition(".")[2][:2]) < 21:
        raise ImportError
    build_cython = True
except ImportError:
    build_cython = False
    # TODO: allow a python-only installation with reduced functionality
    raise EnvironmentError(
        """
        cython could not be found.  Compilation of fast_upfirdn requires Cython
        version >= 0.23.
        Install or upgrade cython via:
        pip install cython --upgrade
        """
    )

extra_compile_args = []
cmdclass = {"build_ext": build_ext}

src_path = os.path.join("fast_upfirdn", "cpu")

# C extensions
ext_upfirdn = Extension(
    "fast_upfirdn.cpu._upfirdn_apply",
    sources=[os.path.join(src_path, "_upfirdn_apply.pyx")],
    language="c",
    extra_compile_args=extra_compile_args,
    include_dirs=[numpy_include],
)

ext_modules = [ext_upfirdn]

c_macros = [("PY_EXTENSION", None)]
cython_macros = []
cythonize_opts = {"language_level": "3"}
if os.environ.get("CYTHON_TRACE"):
    cythonize_opts["linetrace"] = True
    cython_macros.append(("CYTHON_TRACE_NOGIL", 1))

if USE_CYTHON:
    ext_modules = cythonize(ext_modules, compiler_directives=cythonize_opts)

opts = dict(
    name=NAME,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    url=URL,
    download_url=DOWNLOAD_URL,
    license=LICENSE,
    classifiers=CLASSIFIERS,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    platforms=PLATFORMS,
    version=VERSION,
    packages=PACKAGES,
    package_data=PACKAGE_DATA,
    install_requires=REQUIRES,
    python_requires=PYTHON_REQUIRES,
    setup_requires=SETUP_REQUIRES,
    requires=REQUIRES,
    cmdclass=cmdclass,
    ext_modules=ext_modules,
)


if __name__ == "__main__":
    setup(**opts)
